#include <asterisk/lock.h>
#include "astodbc.h"
#include "Vector.h"
#include "String.h"
#include "ODBCQuery.h"
#include "AutoMutex.h"

////////////////////////////////////////////////////////////////////////////////////////////
// types
////////////////////////////////////////////////////////////////////////////////////////////
struct ODBCConnectionEx : public ODBCConnection {
	CString m_last_error_text;
};
struct ODBCQueryEx : public ODBCQuery {
	CString m_last_error_text;
};

////////////////////////////////////////////////////////////////////////////////////////////
// static data
////////////////////////////////////////////////////////////////////////////////////////////
AST_MUTEX_DEFINE_STATIC(_lock);
static Vector<ODBCConnectionEx,50>  _connections;
static Vector<ODBCQueryEx,50>       _queries;
static CString                      _last_error_text;

////////////////////////////////////////////////////////////////////////////////////////////
// static code
////////////////////////////////////////////////////////////////////////////////////////////
int _compare_connections( const void* c1, const ODBCConnectionEx* c2 ) {
	return (int)c1-(int)c2;
}
int _compare_queries( const void* q1, const ODBCQueryEx* q2 ) {
	return (int)q1-(int)q2;
}

////////////////////////////////////////////////////////////////////////////////////////////
// astodbc API
////////////////////////////////////////////////////////////////////////////////////////////
const char* ast_odbc_conn_new( const char* connString, short poolable, short writeable, void** pconn ) {
	AutoMutex autoLock(_lock);
        ODBCConnectionEx* odbcConnection = new ODBCConnectionEx;
        if( odbcConnection ) {
                if( (_last_error_text=odbcConnection->open(connString,poolable,writeable))=="" ) {
                        if( _connections.add(odbcConnection,_connections.bsearch_position(odbcConnection,_compare_connections))>0 ) {
                                *pconn = odbcConnection;
                                return odbcConnection->m_last_error_text = "";
                        }
                        else {
                                _last_error_text = "Unable to add to connection pool";
                        }
                }
                else {
                        _last_error_text = formatString("Cannot open connection (%s)",(const char*)_last_error_text);
                }
                delete odbcConnection;
        }
        else {
                _last_error_text = "Cannot allocate ODBConnection";
        }
        return _last_error_text;
}

const char* ast_odbc_conn_error_text( void* conn ) {
	AutoMutex autoLock(_lock);
        if( _connections.bsearch(conn,_compare_connections) ) {
		ODBCConnectionEx* odbcConnection = (ODBCConnectionEx*)conn;
                return odbcConnection->m_last_error_text = odbcConnection->getErrorText();
        }
	return _last_error_text = "Invalid argument";
}

void* ast_odbc_conn_delete( void* conn ) {
	AutoMutex autoLock(_lock);
	bool found;
	int pos = _connections.bsearch_position((ODBCConnectionEx*)conn,_compare_connections,&found);
	if( found ) {
                delete _connections.remove(pos);
                conn = NULL;
        }
        return conn;
}

const char* ast_odbc_query_new( void* conn, const char* queryString, void** pquery ) {
	AutoMutex autoLock(_lock);
	const char* result = "unknown";
        if( _connections.bsearch(conn,_compare_connections)>=0 ) {
		ODBCConnectionEx* odbcConnection = (ODBCConnectionEx*)conn;
                ODBCQueryEx* odbcQuery = new ODBCQueryEx;
                if( odbcQuery ) {
                        if( (odbcConnection->m_last_error_text=odbcQuery->open(*odbcConnection,queryString))=="" ) {
                                if( _queries.add(odbcQuery,_queries.bsearch_position(odbcQuery,_compare_queries))>0 ) {
                                        *pquery = odbcQuery;
                                        return odbcConnection->m_last_error_text = odbcQuery->m_last_error_text = "";
                                }
				else {
					result = odbcConnection->m_last_error_text = "Unable to add to query pool";
				}
                        }
			else {
				odbcConnection->m_last_error_text = formatString("Unable to open query (%s)",
										 (const char*)odbcConnection->m_last_error_text);
			}
			delete odbcQuery;
                }
                else {
			result = odbcConnection->m_last_error_text = "Unable to allocate ODBCQuery";
                }
        }
        else {
                result = _last_error_text = "Invalid connection argument";
        }
        return result;
}

short ast_odbc_query_eof( void* query ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->eof();
	}
	return 0;
}

const char* ast_odbc_query_fetch( void* query ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->m_last_error_text = odbcQuery->fetch();
	}
	return _last_error_text = "Invalid argument";
}

short ast_odbc_query_next( void* query ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->next();
	}
	return 0;
}

const char* ast_odbc_query_getstring( void* query, const char* fname, const char* defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_string(fname,defaultValue);
	}
	return defaultValue;
}

int ast_odbc_query_getint( void* query, const char* fname, int defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_int(fname,defaultValue);
	}
	return defaultValue;
}

short ast_odbc_query_getbool( void* query, const char* fname, short defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_bool(fname,defaultValue);
	}
	return defaultValue;
}

time_t ast_odbc_query_gettime( void* query, const char* fname, time_t defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_time(fname,defaultValue);
	}
	return defaultValue;
}

const char* ast_odbc_query_getstring_by_ndx( void* query, int fndx, const char* defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_string_by_ndx(fndx,defaultValue);
	}
	return defaultValue;
}

int ast_odbc_query_getint_by_ndx( void* query, int fndx, int defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_int_by_ndx(fndx,defaultValue);
	}
	return defaultValue;
}

short ast_odbc_query_getbool_by_ndx( void* query, int fndx, short defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_bool_by_ndx(fndx,defaultValue);
	}
	return defaultValue;
}

time_t ast_odbc_query_gettime_by_ndx( void* query, int fndx, time_t defaultValue ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_time_by_ndx(fndx,defaultValue);
	}
	return defaultValue;
}

const char* ast_odbc_query_error_text( void* query ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->m_last_error_text = odbcQuery->getErrorText();
	}
	return _last_error_text = "Invalid argument";
}

short ast_odbc_query_got_column( void* query, const char* fname ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->got_column(fname);
	}
	return false;
}

int ast_odbc_query_get_column_count( void* query ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_column_count();
	}
	return -1;
}

const char* ast_odbc_query_get_column_name( void* query, int ndx ) {
	AutoMutex autoLock(_lock);
	if( _queries.bsearch((ODBCQueryEx*)query,_compare_queries) ) {
		ODBCQueryEx* odbcQuery = (ODBCQueryEx*)query;
		return odbcQuery->get_column_name(ndx);
	}
	return NULL;
}


void* ast_odbc_query_delete( void* query ) {
	AutoMutex autoLock(_lock);
	bool found;
	int pos = _queries.bsearch_position((ODBCQueryEx*)query,_compare_queries,&found);
	if( found ) {
                delete _queries.remove(pos);
                query = NULL;
        }
        return query;
}
